@@ -1,10 +1,11 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 from decimal import Decimal
 from collections import namedtuple
 from itertools import groupby
 
-from sql.aggregate import Sum
+from sql import Null
+from sql.aggregate import Max, Sum
 
 from trytond.model import Workflow, ModelView, ModelSQL, fields
 from trytond.pyson import Eval, If, Bool
@@ -12,9 +13,8 @@ from trytond.transaction import Transaction
 from trytond import backend
 from trytond.pool import Pool
 from trytond.modules.company import CompanyReport
-from trytond.tools import reduce_ids
 
-__all__ = ['Statement', 'Line', 'StatementReport']
+__all__ = ['Statement', 'Line', 'LineGroup', 'StatementReport']
 
 _STATES = {'readonly': Eval('state') != 'draft'}
 _DEPENDS = ['state']
@@ -41,7 +41,7 @@ _NUMBER_STATES.update({
 _NUMBER_DEPENDS = _DEPENDS + ['validation']
 
 
-class Null(object):
+class Unequal(object):
     "Always different"
 
     def __eq__(self, other):
@@ -67,12 +67,12 @@ class Statement(Workflow, ModelSQL, ModelView):
     journal = fields.Many2One('account.statement.journal', 'Journal',
         required=True, select=True,
         domain=[
-            ('company', '=', Eval('context', {}).get('company', -1)),
+            ('company', '=', Eval('company', -1)),
             ],
         states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
             },
-        depends=['state'])
+        depends=['state', 'company'])
     currency_digits = fields.Function(fields.Integer('Currency Digits'),
         'on_change_with_currency_digits')
     date = fields.Date('Date', required=True, states=_STATES, depends=_DEPENDS,
@@ -120,6 +120,8 @@ class Statement(Workflow, ModelSQL, ModelView):
                     'deletion.'),
                 'paid_invoice_draft_statement': ('There are paid invoices on '
                     'draft statements.'),
+                'debit_credit_account_statement_journal': ('Please provide '
+                    'debit and credit account on statement journal "%s".'),
                 })
         cls._transitions |= set((
                 ('draft', 'validated'),
@@ -177,7 +179,7 @@ class Statement(Workflow, ModelSQL, ModelView):
         # Migration from 3.2: add required name
         cursor.execute(*sql_table.update([sql_table.name],
                 [sql_table.id.cast(cls.name.sql_type().base)],
-                where=sql_table.name == None))
+                where=sql_table.name == Null))
 
     @staticmethod
     def default_company():
@@ -202,9 +204,8 @@ class Statement(Workflow, ModelSQL, ModelView):
 
     @fields.depends('journal', 'state', 'lines')
     def on_change_journal(self):
-        res = {}
         if not self.journal:
-            return res
+            return
 
         statements = self.search([
                 ('journal', '=', self.journal.id),
@@ -212,11 +213,10 @@ class Statement(Workflow, ModelSQL, ModelView):
                 ('date', 'DESC'),
                 ], limit=1)
         if not statements:
-            return res
+            return
 
         statement, = statements
-        res['start_balance'] = statement.end_balance
-        return res
+        self.start_balance = statement.end_balance
 
     @fields.depends('journal')
     def on_change_with_currency_digits(self, name=None):
@@ -240,9 +240,6 @@ class Statement(Workflow, ModelSQL, ModelView):
         pool = Pool()
         Currency = pool.get('currency.currency')
         Line = pool.get('account.statement.line')
-        res = {
-            'lines': {},
-            }
         if self.journal and self.lines:
             invoices = set()
             for line in self.lines:
@@ -259,52 +256,36 @@ class Statement(Workflow, ModelSQL, ModelView):
                         Currency.compute(invoice.currency,
                             invoice.amount_to_pay, self.journal.currency))
 
+            lines = list(self.lines)
             line_offset = 0
             for index, line in enumerate(self.lines or []):
                 if getattr(line, 'invoice', None) and line.id:
                     amount_to_pay = invoice_id2amount_to_pay[line.invoice.id]
-                    res['lines'].setdefault('update', [])
                     if (not self.journal.currency.is_zero(amount_to_pay)
                             and getattr(line, 'amount', None)
                             and (line.amount >= 0) == (amount_to_pay <= 0)):
-                        res['lines']['update'].append({
-                                'id': line.id,
-                                'amount': (amount_to_pay.copy_sign(line.amount)
-                                    if abs(line.amount) > abs(amount_to_pay)
-                                    else line.amount)
-                                })
                         if abs(line.amount) > abs(amount_to_pay):
-                            res['lines'].setdefault('add', [])
-                            vals = {}
+                            new_line = Line()
                             for field_name, field in Line._fields.iteritems():
-                                try:
-                                    value = getattr(line, field_name)
-                                except AttributeError:
+                                if field_name == 'id':
                                     continue
-                                if (value and field._type in ('many2one',
-                                            'one2one')):
-                                    vals[field_name] = value.id
-                                    if value.id >= 0:
-                                        vals[field_name + '.rec_name'] = \
-                                            value.rec_name
-                                else:
-                                    vals[field_name] = value
-                            del vals['id']
-                            vals['amount'] = line.amount + amount_to_pay
-                            vals['invoice'] = None
-                            del vals['invoice.rec_name']
+                                try:
+                                    setattr(new_line, field_name,
+                                        getattr(line, field_name))
+                                except AttributeError:
+                                    pass
+                            new_line.amount = line.amount + amount_to_pay
+                            new_line.invoice = None
                             line_offset += 1
-                            res['lines']['add'].append((index + line_offset, vals))
+                            lines.insert(index + line_offset, new_line)
                             invoice_id2amount_to_pay[line.invoice.id] = 0
+                            line.amount = amount_to_pay.copy_sign(line.amount)
                         else:
                             invoice_id2amount_to_pay[line.invoice.id] = (
                                 line.amount + amount_to_pay)
                     else:
-                        res['lines']['update'].append({
-                                'id': line.id,
-                                'invoice': None,
-                                })
-        return res
+                        line.invoice = None
+            self.lines = lines
 
     @fields.depends('journal')
     def on_change_with_validation(self, name=None):
@@ -313,7 +294,7 @@ class Statement(Workflow, ModelSQL, ModelView):
 
     def _group_key(self, line):
         key = (
-            ('number', line.number or Null()),
+            ('number', line.number or Unequal()),
             ('date', line.date),
             ('party', line.party),
             )
@@ -405,8 +386,9 @@ class Statement(Workflow, ModelSQL, ModelView):
 
         for statement in statements:
             getattr(statement, 'validate_%s' % statement.validation)()
-            for line in statement.lines:
-                line.create_move()
+
+        cls.create_move(statements)
+
         cls.write(statements, {
                 'state': 'validated',
                 })
@@ -421,6 +403,103 @@ class Statement(Workflow, ModelSQL, ModelView):
                     'invoice': None,
                     })
 
+    @classmethod
+    def create_move(cls, statements):
+        '''Create move for the statements and try to reconcile the lines.
+        Returns the list of move, statement and lines
+        '''
+        pool = Pool()
+        Line = pool.get('account.statement.line')
+        Move = pool.get('account.move')
+        MoveLine = pool.get('account.move.line')
+
+        moves = []
+        for statement in statements:
+            for key, lines in groupby(
+                    statement.lines, key=statement._group_key):
+                lines = list(lines)
+                key = dict(key)
+                move = statement._get_move(key)
+                moves.append((move, statement, lines))
+
+        Move.save([m for m, _, _ in moves])
+
+        to_write = []
+        for move, _, lines in moves:
+            to_write.append(lines)
+            to_write.append({
+                    'move': move.id,
+                    })
+        Line.write(*to_write)
+
+        move_lines = []
+        for move, statement, lines in moves:
+            amount = 0
+            amount_second_currency = 0
+            for line in lines:
+                move_line = line.get_move_line()
+                move_line.move = move
+                amount += move_line.debit - move_line.credit
+                if move_line.amount_second_currency:
+                    amount_second_currency += move_line.amount_second_currency
+                move_lines.append((move_line, line))
+
+            move_line = statement._get_move_line(
+                amount, amount_second_currency)
+            move_line.move = move
+            move_lines.append((move_line, None))
+
+        MoveLine.save([l for l, _ in move_lines])
+
+        for move_line, line in move_lines:
+            if line:
+                line.reconcile(move_line)
+        return moves
+
+    def _get_move(self, key):
+        'Return Move for the grouping key'
+        pool = Pool()
+        Move = pool.get('account.move')
+        Period = pool.get('account.period')
+
+        period_id = Period.find(self.company.id, date=key['date'])
+        return Move(
+            period=period_id,
+            journal=self.journal.journal,
+            date=key['date'],
+            origin=self,
+            company=self.company,
+            )
+
+    def _get_move_line(self, amount, amount_second_currency):
+        'Return counterpart Move Line for the amount'
+        pool = Pool()
+        MoveLine = pool.get('account.move.line')
+
+        if amount < 0:
+            account = self.journal.journal.debit_account
+        else:
+            account = self.journal.journal.credit_account
+
+        if not account:
+            self.raise_user_error('debit_credit_account_statement_journal',
+                (self.journal.rec_name,))
+
+        if self.journal.currency != self.company.currency:
+            second_currency = self.journal.currency
+            amount_second_currency *= -1
+        else:
+            second_currency = None
+            amount_second_currency = None
+
+        return MoveLine(
+            debit=abs(amount) if amount < 0 else 0,
+            credit=abs(amount) if amount > 0 else 0,
+            account=account,
+            second_currency=second_currency,
+            amount_second_currency=amount_second_currency,
+            )
+
     @classmethod
     @ModelView.button
     @Workflow.transition('posted')
@@ -457,7 +536,10 @@ class Line(ModelSQL, ModelView):
             ('kind', '!=', 'view'),
             ])
     description = fields.Char('Description')
-    move = fields.Many2One('account.move', 'Account Move', readonly=True)
+    move = fields.Many2One('account.move', 'Account Move', readonly=True,
+        domain=[
+            ('company', '=', Eval('_parent_statement', {}).get('company', -1)),
+            ])
     invoice = fields.Many2One('account.invoice', 'Invoice',
         domain=[
             If(Bool(Eval('party')), [('party', '=', Eval('party'))], []),
@@ -473,11 +555,6 @@ class Line(ModelSQL, ModelView):
         super(Line, cls).__setup__()
         cls._order.insert(0, ('sequence', 'ASC'))
         cls._error_messages.update({
-                'debit_credit_account_statement_journal': ('Please provide '
-                    'debit and credit account on statement journal "%s".'),
-                'same_debit_credit_account': ('Account "%(account)s" in '
-                    'statement line "%(line)s" is the same as the one '
-                    'configured as credit or debit on journal "%(journal)s".'),
                 'amount_greater_invoice_amount_to_pay': ('Amount "%s" is '
                     'greater than the amount to pay of invoice.'),
                 })
@@ -489,7 +566,7 @@ class Line(ModelSQL, ModelView):
     @staticmethod
     def order_sequence(tables):
         table, _ = tables[None]
-        return [table.sequence == None, table.sequence]
+        return [table.sequence == Null, table.sequence]
 
     @staticmethod
     def default_amount():
@@ -497,31 +574,24 @@ class Line(ModelSQL, ModelView):
 
     @fields.depends('amount', 'party', 'invoice')
     def on_change_party(self):
-        res = {}
-
         if self.party:
             if self.amount:
                 if self.amount > Decimal("0.0"):
-                    account = self.party.account_receivable
+                    self.account = self.party.account_receivable
                 else:
-                    account = self.party.account_payable
-                res['account'] = account.id
-                res['account.rec_name'] = account.rec_name
+                    self.account = self.party.account_payable
 
         if self.invoice:
             if self.party:
                 if self.invoice.party != self.party:
-                    res['invoice'] = None
+                    self.invoice = None
             else:
-                res['invoice'] = None
-        return res
+                self.invoice = None
 
     @fields.depends('amount', 'party', 'account', 'invoice',
         '_parent_statement.journal')
     def on_change_amount(self):
         Currency = Pool().get('currency.currency')
-        res = {}
-
         if self.party:
             if self.account and self.account not in (
                     self.party.account_receivable, self.party.account_payable):
@@ -529,11 +599,9 @@ class Line(ModelSQL, ModelView):
                 pass
             elif self.amount:
                 if self.amount > Decimal("0.0"):
-                    account = self.party.account_receivable
+                    self.account = self.party.account_receivable
                 else:
-                    account = self.party.account_payable
-                res['account'] = account.id
-                res['account.rec_name'] = account.rec_name
+                    self.account = self.party.account_payable
         if self.invoice:
             if self.amount and self.statement and self.statement.journal:
                 invoice = self.invoice
@@ -542,34 +610,26 @@ class Line(ModelSQL, ModelView):
                     amount_to_pay = Currency.compute(invoice.currency,
                         invoice.amount_to_pay, journal.currency)
                 if abs(self.amount) > amount_to_pay:
-                    res['invoice'] = None
+                    self.invoice = None
             else:
-                res['invoice'] = None
-        return res
+                self.invoice = None
 
     @fields.depends('account', 'invoice')
     def on_change_account(self):
-        res = {}
-
         if self.invoice:
             if self.account:
                 if self.invoice.account != self.account:
-                    res['invoice'] = None
+                    self.invoice = None
             else:
-                res['invoice'] = None
-        return res
+                self.invoice = None
 
     @fields.depends('party', 'account', 'invoice')
     def on_change_invoice(self):
-        changes = {}
         if self.invoice:
             if not self.party:
-                changes['party'] = self.invoice.party.id
-                changes['party.rec_name'] = self.invoice.party.rec_name
+                self.party = self.invoice.party
             if not self.account:
-                changes['account'] = self.invoice.account.id
-                changes['account.rec_name'] = self.invoice.account.rec_name
-        return changes
+                self.account = self.invoice.account
 
     def get_rec_name(self, name):
         return self.statement.rec_name
@@ -587,36 +647,12 @@ class Line(ModelSQL, ModelView):
         default.setdefault('invoice', None)
         return super(Line, cls).copy(lines, default=default)
 
-    def create_move(self):
-        '''
-        Create move for the statement line and return move if created.
-        '''
+    def reconcile(self, move_line):
         pool = Pool()
-        Move = pool.get('account.move')
-        Period = pool.get('account.period')
-        Invoice = pool.get('account.invoice')
         Currency = pool.get('currency.currency')
-        MoveLine = pool.get('account.move.line')
         Lang = pool.get('ir.lang')
-
-        if self.move:
-            return
-
-        period_id = Period.find(self.statement.company.id, date=self.date)
-
-        move_lines = self._get_move_lines()
-        move = Move(
-            period=period_id,
-            journal=self.statement.journal.journal,
-            date=self.date,
-            origin=self,
-            lines=move_lines,
-            )
-        move.save()
-
-        self.write([self], {
-                'move': move.id,
-                })
+        Invoice = pool.get('account.invoice')
+        MoveLine = pool.get('account.move.line')
 
         if self.invoice:
             with Transaction().set_context(date=self.invoice.currency_date):
@@ -641,30 +677,28 @@ class Line(ModelSQL, ModelView):
             reconcile_lines = self.invoice.get_reconcile_lines_for_amount(
                 amount)
 
-            for move_line in move.lines:
-                if move_line.account == self.invoice.account:
-                    Invoice.write([self.invoice], {
-                            'payment_lines': [('add', [move_line.id])],
-                            })
-                    break
+            assert move_line.account == self.invoice.account
+
+            Invoice.write([self.invoice], {
+                    'payment_lines': [('add', [move_line.id])],
+                    })
             if reconcile_lines[1] == Decimal('0.0'):
                 lines = reconcile_lines[0] + [move_line]
                 MoveLine.reconcile(lines)
-        return move
 
     @classmethod
     def post_move(cls, lines):
         Move = Pool().get('account.move')
-        Move.post([l.move for l in lines if l.move])
+        Move.post(list({l.move for l in lines if l.move}))
 
     @classmethod
     def delete_move(cls, lines):
         Move = Pool().get('account.move')
-        Move.delete([l.move for l in lines if l.move])
+        Move.delete(list({l.move for l in lines if l.move}))
 
-    def _get_move_lines(self):
+    def get_move_line(self):
         '''
-        Return the move lines for the statement line
+        Return the move line for the statement line
         '''
         pool = Pool()
         MoveLine = pool.get('account.move.line')
@@ -675,46 +709,93 @@ class Line(ModelSQL, ModelView):
                 self.amount, self.statement.company.currency)
         if self.statement.journal.currency != self.statement.company.currency:
             second_currency = self.statement.journal.currency.id
-            amount_second_currency = abs(self.amount)
+            amount_second_currency = -self.amount
         else:
             amount_second_currency = None
             second_currency = None
 
-        move_lines = []
-        move_lines.append(MoveLine(
-                description=self.description,
-                debit=amount < zero and -amount or zero,
-                credit=amount >= zero and amount or zero,
-                account=self.account,
-                party=self.party if self.account.party_required else None,
-                second_currency=second_currency,
-                amount_second_currency=amount_second_currency,
-                ))
+        return MoveLine(
+            description=self.description,
+            debit=amount < zero and -amount or zero,
+            credit=amount >= zero and amount or zero,
+            account=self.account,
+            party=self.party if self.account.party_required else None,
+            second_currency=second_currency,
+            amount_second_currency=amount_second_currency,
+            )
 
-        journal = self.statement.journal.journal
-        if self.amount >= zero:
-            account = journal.debit_account
-        else:
-            account = journal.credit_account
-        if not account:
-            self.raise_user_error('debit_credit_account_statement_journal',
-                (journal.rec_name,))
-        if self.account == account:
-            self.raise_user_error('same_debit_credit_account', {
-                    'account': self.account.rec_name,
-                    'line': self.rec_name,
-                    'journal': journal,
-                    })
-        move_lines.append(MoveLine(
-                description=self.description,
-                debit=amount >= zero and amount or zero,
-                credit=amount < zero and -amount or zero,
-                account=account,
-                party=self.party if account.party_required else None,
-                second_currency=second_currency,
-                amount_second_currency=amount_second_currency,
-                ))
-        return move_lines
+
+class LineGroup(ModelSQL, ModelView):
+    'Account Statement Line Group'
+    __name__ = 'account.statement.line.group'
+    _rec_name = 'number'
+    statement = fields.Many2One('account.statement', 'Statement')
+    journal = fields.Function(fields.Many2One('account.statement.journal',
+            'Journal'), 'get_journal', searcher='search_journal')
+    number = fields.Char('Number')
+    date = fields.Date('Date')
+    amount = fields.Numeric('Amount',
+        digits=(16, Eval('currency_digits', 2)),
+        depends=['currency_digits'])
+    currency = fields.Function(fields.Many2One('currency.currency',
+            'Currency'), 'get_currency')
+    currency_digits = fields.Function(fields.Integer('Currency Digits'),
+        'get_currency_digits')
+    party = fields.Many2One('party.party', 'Party')
+    move = fields.Many2One('account.move', 'Move')
+
+    @classmethod
+    def __setup__(cls):
+        super(LineGroup, cls).__setup__()
+        cls._order.insert(0, ('date', 'DESC'))
+
+    @classmethod
+    def _grouped_columns(cls, line):
+        return [
+            Max(line.statement).as_('statement'),
+            Max(line.number).as_('number'),
+            Max(line.date).as_('date'),
+            Sum(line.amount).as_('amount'),
+            Max(line.party).as_('party'),
+            ]
+
+    @classmethod
+    def table_query(cls):
+        pool = Pool()
+        Move = pool.get('account.move')
+        Line = pool.get('account.statement.line')
+        move = Move.__table__()
+        line = Line.__table__()
+
+        std_columns = [
+            move.id,
+            move.create_uid,
+            move.create_date,
+            move.write_uid,
+            move.write_date,
+            ]
+
+        columns = (std_columns + [move.id.as_('move')]
+            + cls._grouped_columns(line))
+        return move.join(line,
+            condition=move.id == line.move
+            ).select(*columns,
+                where=move.origin.ilike(Statement.__name__ + ',%'),
+                group_by=std_columns + [move.id]
+                )
+
+    def get_journal(self, name):
+        return self.statement.journal.id
+
+    @classmethod
+    def search_journal(cls, name, clause):
+        return [('statement.journal',) + tuple(clause[1:])]
+
+    def get_currency(self, name):
+        return self.statement.journal.currency.id
+
+    def get_currency_digits(self, name):
+        return self.statement.journal.currency.digits
 
 
 class StatementReport(CompanyReport):
